package it.eng.spago.cms.init;

import it.eng.javax.jcr.Node;
import it.eng.javax.jcr.Session;
import it.eng.spago.base.Constants;
import it.eng.spago.base.SourceBean;
import it.eng.spago.cms.ISessionFactory;
import it.eng.spago.cms.exec.operations.SetOperation;
import it.eng.spago.cms.pool.CMSPool;
import it.eng.spago.cms.pool.CMSPoolImpl;
import it.eng.spago.cms.util.Path;
import it.eng.spago.cms.util.constants.ModuleConstants;
import it.eng.spago.cms.util.constants.OperationsConstants;
import it.eng.spago.init.InitializerIFace;
import it.eng.spago.tracing.TracerSingleton;

import java.util.Iterator;
import java.util.List;
import java.util.Vector;


/**
 * 
 * This class initialize the Content Management System 
 * using some configuration parameters.
 * The class is call from Spago Framework during the 
 * Application Start-Up.
 * Create the repository and all objects necessary for using it   
 * 
 */
public class CMSInitializer implements InitializerIFace {
	
	/** 
	 * SourceBean that contains the configuration parameters
	 */
	private SourceBean _config = null;
	/**
	 * System prefix for all node or property set by the system
	 */
	private final String PREFIX_APP_SYS = "appSys";
	/**
	 * Namespace associated with the System prefix
	 */
	private final String VALUE_PREFIX_APP_SYS = "http://it.eng.spago.contentRepository.system";
	/**
	 * User Prefix for properties set by the user
	 */
	private final String PREFIX_APP_USER = "appUsr";
    /**
     * Namespace associated with the user prefix
     */
	private final String VALUE_PREFIX_APP_USER = "http://it.eng.spago.contentRepository.user";


	
	
	
	
	
	
	/**
	 * Create and initialize all the repositories define
	 * in the configuration SourceBean
	 * 
	 * @param config, SourceBean containg params configuration
	 * for each repository to create
	 */
    public void init(SourceBean config) {
    	
    	TracerSingleton.log(Constants.NOME_MODULO, TracerSingleton.DEBUG, "ContentRepositoryInitializer::init: config", config);
        _config = config;
        
        // control for configuration sourceBean not null
        if(config == null) {
        	TracerSingleton.log(Constants.NOME_MODULO, TracerSingleton.CRITICAL,
            "ContentRepositoryInitializer::init:configuration sourceBean null \n");
            System.out.println("ContentRepositoryInitializer::init: configuration sourceBean null\n");
            return;
        }
                
        // get List of sourceBean configuration descriptor for all repository 
        List repositories = config.getAttributeAsList("CONTENTREPOSITORY");  
        
        // control if at least one repository is configured
        if(repositories.size() == 0) {
        	TracerSingleton.log(Constants.NOME_MODULO, TracerSingleton.WARNING,
            "ContentRepositoryInitializer::init: no configured repository \n");
            System.out.println("ContentRepositoryInitializer::init: no configured repository \n");
            return;
        } 
        
        // get the default name repository
        String defaultRep = "";
        Object objDefRep = config.getAttribute("DEFAULTREPOSITORY.name");
        if(objDefRep != null) {
        	defaultRep = objDefRep.toString().trim();
        }
        else {
        	TracerSingleton.log(Constants.NOME_MODULO, TracerSingleton.WARNING,
            "ContentRepositoryInitializer::init: default repository not set \n");
            System.out.println("ContentRepositoryInitializer::init: rdefault repository not set \n");
            return;
        }
        
        // control if default repository is one of the configured repositories
        boolean find = false;
        for(int i=0; i<repositories.size(); i++) {
        	Object objRepBean = repositories.get(i);
        	if((objRepBean == null)  ||  !(objRepBean instanceof SourceBean) ) {
        		continue;
        	}
        	SourceBean repositoryBean = (SourceBean)objRepBean;
        	Object objRepName = repositoryBean.getAttribute("NAME");
        	if(objRepName == null) {
        		continue;
        	}
        	String repositoryName = objRepName.toString().trim();
        	if(defaultRep.equals(repositoryName)) {
        		find = true; 
        	}
        }
        if(!find){
        	TracerSingleton.log(Constants.NOME_MODULO, TracerSingleton.WARNING,
            "ContentRepositoryInitializer::init: default repository name doesn't exists \n");
            System.out.println("ContentRepositoryInitializer::init: default repository name doesn't exists \n");
            return; 	
        }
        
        // registry the name of default repository
        CMSManager.setDefaultRepository(defaultRep);
        
        
        // FOR ALL REPOSITORIES
        Vector nameRepRegistered = new Vector();
        for (int i = 0; i < repositories.size(); i++) {
            try {
            	ISessionFactory contentRepository = null;
                CMSPool pool  = null;
                                
            	//get configuration sourceBean of the repository
                SourceBean repositoryBean = (SourceBean)repositories.get(i);
                
                // get credentails for create JCR Session
                String userForConn = null;
                String passwordForConn = null;
                String workspaceForConn = null;
                Object userForConnObj = repositoryBean.getAttribute("CONNECTIONDATA.user");
                if(userForConnObj != null) {
                	userForConn = userForConnObj.toString();
                }
                Object passwordForConObj = repositoryBean.getAttribute("CONNECTIONDATA.password");
                if(passwordForConObj != null) {
                	passwordForConn = passwordForConObj.toString();
                }
                Object workspaceForConnObj = (String)repositoryBean.getAttribute("CONNECTIONDATA.workspace");
                if(workspaceForConnObj != null) {
                	workspaceForConn = workspaceForConnObj.toString();
                }
                
                
                // get the class to instantiate and repository name
                String repositoryClass = ((String)repositoryBean.getAttribute("CLASS")).trim();
                String repositoryName = ((String)repositoryBean.getAttribute("NAME")).trim();
                // se il nome del repository � gi� registrato sollevo un eccezione (catturata dal cath finale) 
                if(nameRepRegistered.contains(repositoryName)) {
                	TracerSingleton.log(Constants.NOME_MODULO, TracerSingleton.WARNING,
                    "ContentRepositoryInitializer::init: repository name already in use : "+repositoryName+"\n");
                    throw new Exception("repository name already in use : "+repositoryName);	
                }
                try {
                	 contentRepository = (ISessionFactory)Class.forName(repositoryClass).newInstance();
                }
                catch(Exception e) {
                	 TracerSingleton.log(Constants.NOME_MODULO, TracerSingleton.WARNING,
                     "ContentRepositoryInitializer::init: impossible to create the repository "+repositoryName+"\n"+e);
                     throw e;
                }
                // creation of the application namespace
                Session sess = null;
                try {
                	sess = contentRepository.getSession(userForConn, passwordForConn, workspaceForConn);
                	String [] prefxs = sess.getWorkspace().getNamespaceRegistry().getPrefixes();
	                boolean notRegistrySysPre = true;
	                boolean notRegistryUsrPre = true;
			        for(int j=0; j<prefxs.length; j++) {
			        	if(prefxs[j].equals(PREFIX_APP_SYS)) {
			        		notRegistrySysPre = false;
			        	}
			        	if(prefxs[j].equals(PREFIX_APP_USER)) {
			        		notRegistryUsrPre = false;
			        	}
			        }
			        if(notRegistrySysPre) {
			            sess.getWorkspace().getNamespaceRegistry().registerNamespace(PREFIX_APP_SYS, VALUE_PREFIX_APP_SYS);
			        }
			        if(notRegistryUsrPre) {
			            sess.getWorkspace().getNamespaceRegistry().registerNamespace(PREFIX_APP_USER, VALUE_PREFIX_APP_USER);
			        }
			        sess.logout();
                }
                catch(Exception e) {
                	 if(sess != null) {
                	 	sess.logout();
                	 }
                	 TracerSingleton.log(Constants.NOME_MODULO, TracerSingleton.WARNING,
                     "ContentRepositoryInitializer::init: errore during namespaces registration " +
                     "per il repository "+repositoryName+"\n"+e);
                     throw e;
                }
                // creation of object pool
                try {
                	 SourceBean configPool = (SourceBean)repositoryBean.getAttribute("POOLCONFIGURATION");
                     pool  = new CMSPoolImpl(contentRepository, configPool, repositoryName, 
                     		                               userForConn, passwordForConn, workspaceForConn, 
														   PREFIX_APP_SYS, PREFIX_APP_USER);
                }
                catch(Exception e) {
                	 TracerSingleton.log(Constants.NOME_MODULO, TracerSingleton.WARNING,
                     "ContentRepositoryInitializer::init: impossible to create connection pool for " +
                     "il repository "+ repositoryName + "\n" + e);
                     throw e;
                }
                try {
                	CMSManager.registerRepository(pool, repositoryName);
                	nameRepRegistered.add(repositoryName);
                }
                catch(Exception e) {
                	 TracerSingleton.log(Constants.NOME_MODULO, TracerSingleton.WARNING,
                     "ContentRepositoryInitializer::init: impossible to registry connection pool of " +
                     "repository " + repositoryName + " \n " + e);
                     throw e;
                }
                
                

                
                // creation of the initial structure
                // ATTENTION: the following code assume that into the configuraion file the nodes to create 
                // are defined in gerarchy order. So for each path only the last element has to be created.
                // Example: /node1/node2/node3 this path assume that only the node3 not exist.
                // In future realeses this behaviour will be modified in ordet to add more flexibility.
                sess = null;
                try {
                	sess = contentRepository.getSession(userForConn, passwordForConn, workspaceForConn);
                	Node rootNode = sess.getRootNode();
                    List nodesInitStruct = repositoryBean.getAttributeAsList("INITIALSTRUCTURE.NODE");
                    Iterator iterNIS = nodesInitStruct.iterator();
                    String pathStr = "";
                    while(iterNIS.hasNext()) {
                    	try {
                    		SourceBean pathSB = (SourceBean)iterNIS.next();
                    		pathStr = (String)pathSB.getAttribute("path");
                    		Path pathObj =  Path.create(pathStr);
                    		
                    		try {
                    			rootNode.getNode(pathObj.getRootRelativePathStr());
                    			continue;
                    		} catch (Exception e) {
                    			// ignore : node doesn't exist and must be created 
                    		}
                    		
                    		Path PathAncestorObj = pathObj.getPathAncestor(1);
                    		String nameNewNode = pathObj.getNameLastElement();
                    		Node parentNode = rootNode;
                    		if(!PathAncestorObj.getAbsPathStr().equals("/")) {
                    			parentNode = rootNode.getNode(PathAncestorObj.getRootRelativePathStr());
                    		}
                			Node newNode = parentNode.addNode(nameNewNode, "nt:unstructured");
                			newNode.addMixin("mix:versionable");
                			newNode.setProperty(PREFIX_APP_SYS + ":"+OperationsConstants.TYPE, SetOperation.TYPE_CONTAINER);
                			parentNode.save();
 
                    	}  catch (Exception e) {
    	                	TracerSingleton.log(ModuleConstants.NAME_MODULE, TracerSingleton.WARNING,
    	                    "ContentRepositoryInitializer::init: cannot create a path of the  " +
    	                            "initial structure: " + pathStr + " \n " + e);
    	                            throw e;
    	                }
                    }
                } catch (Exception e) {
               	 	if(sess != null) {
               	 		sess.logout();
               	 	}
               	 	TracerSingleton.log(ModuleConstants.NAME_MODULE, TracerSingleton.WARNING,
               	 			"ContentRepositoryInitializer::init: errore during the initial structure creation", e );
                    throw e;
                }
                
                // creation of the initial structure
//                CMSConnection con = pool.getConnection();
//                sess = contentRepository.getSession(userForConn, passwordForConn, workspaceForConn);
//                OperationExecutor executor = OperationExecutorManager.getOperationExecutor();
//                List nodes = repositoryBean.getAttributeAsList("INITIALSTRUCTURE.NODE");
//                Iterator iter = nodes.iterator();
//                String pathStr = "";
//                while(iter.hasNext()) {
//                	try {
//	                	SourceBean pathSB = (SourceBean)iter.next();
//	                	pathStr = (String)pathSB.getAttribute("path");
//	                	if(!sess.getRootNode().hasNode(pathStr.substring(1))) {
//	                		SetOperation setOp = OperationBuilder.buildSetOperation();
//	                		setOp.setPath(pathStr);
//	                		setOp.setType(OperationsConstants.CONTAINER);
//	                		//SourceBean propsSB = new SourceBean("PROPERTIES");
//	                		//List props = pathSB.getAttributeAsList("PROPERTY");
//	                	    //Iterator propIter = props.iterator();
//	                	    //while(propIter.hasNext()) {
//	                	    //	SourceBean propSB = (SourceBean)propIter.next();           	    	
//	                	    //	propsSB.setAttribute(propSB);
//	                	    //}
//	                	    //opSB.setAttribute(propsSB);
//	                	    executor.setObject(con, setOp, false);
//	                	}
//                	}	
//	                catch (Exception e) {
//	                	TracerSingleton.log(Constants.NOME_MODULO, TracerSingleton.WARNING,
//	                    "ContentRepositoryInitializer::init: impossible to create a path of the  " +
//	                            "initial structure: " + pathStr + " \n " + e);
//	                            throw e;
//	                }
//                }
//                sess.logout();
//                con.close();
                
                
            } 

            catch (Exception ex) {
                System.out.println("ContentRepositoryInitializer::init: error during the creation " +
                		" / of a Repository \n" + ex);
            } 
        }// for repositories        
    }


    /**
     * return the SourceBean contains the SourceBean configuration
     * of the repository.
     */
	public SourceBean getConfig() {
		return _config;
	}

}
